
def delete_letter_to_end(self, line, letter) -> None:
    """
    Deletes within a line from a letter by index to the end of the line.
    """
    self.line_string_list[line] = self.line_string_list[line][:letter]


def delete_start_to_letter(self, line, letter) -> None:
    """
    Deletes within a line from the start of the line to a letter by index..
    """
    self.line_string_list[line] = self.line_string_list[line][letter:]


def delete_letter_to_letter(self, line, letter_start, letter_end) -> None:
    """
    Deletes within a line from a letter to a letter by index..
    """
    self.line_string_list[line] = self.line_string_list[line][:letter_start] + \
            self.line_string_list[line][letter_end:]


def delete_entire_line(self, line) -> None:
    """
    Deletes an entire line
    """
    self.line_string_list.pop(line)
    self.maxLines -= 1


def get_entire_line(self, line_index) -> str:
    return self.line_string_list[line_index]


def get_line_from_start_to_char(self, line_index, char_index) -> str:
    return self.line_string_list[line_index][0:char_index]


def get_line_from_char_to_end(self, line_index, char_index) -> str:
    return self.line_string_list[line_index][char_index:]


def get_line_from_char_to_char(self, line_index, char1, char2) -> str:
    if char1 < char2:
        return self.line_string_list[line_index][char1:char2]
    else:
        return self.line_string_list[line_index][char2:char1]


import pygame


def get_rect_coord_from_mouse(self, mouse_x, mouse_y) -> (int, int):
    """
    Return x and y pixel-coordinates for the position of the mouse.
    """
    line = self.get_line_index(mouse_y)
    letter = self.get_letter_index(mouse_x)
    return self.get_rect_coord_from_indizes(line, letter)


def get_rect_coord_from_indizes(self, line, letter) -> (int, int):
    """
    Return x and y pixel-coordinates for line and letter by index.
    """
    line_coord = self.editor_offset_Y + (self.line_gap * (line - self.showStartLine))
    letter_coord = self.xline_start + (letter * self.letter_size_X)
    return letter_coord, line_coord


def render_background_coloring(self) -> None:
    """
    Renders background color of the text area.
    """
    bg_left = self.editor_offset_X + self.lineNumberWidth
    bg_top = self.editor_offset_Y
    bg_width = self.textAreaWidth - self.lineNumberWidth
    bg_height = self.textAreaHeight
    pygame.draw.rect(self.screen, self.codingBackgroundColor, (bg_left, bg_top, bg_width, bg_height))


def render_line_numbers(self) -> None:
    """
    While background rendering is done for all "line-slots" (to overpaint remaining "old" numbers without lines)
    we render line-numbers only for existing string-lines.
    """
    if self.displayLineNumbers and self.rerenderLineNumbers:
        self.rerenderLineNumbers = False
        line_numbers_Y = self.editor_offset_Y  # init for first line
        for x in range(self.showStartLine, self.showStartLine + self.showable_line_numbers_in_editor):

            # background
            pygame.draw.rect(self.screen, self.lineNumberBackgroundColor,
                             (self.editor_offset_X, line_numbers_Y, self.lineNumberWidth, 17))
            # line number
            if x < self.get_showable_lines():
                # x + 1 in order to start with line 1 (only display, logical it's the 0th item in the list
                text = self.courier_font.render(str(x + 1).zfill(2), 1, self.lineNumberColor)
                self.screen.blit(text, (self.editor_offset_X + 5, line_numbers_Y + 3))  # center of bg block

            line_numbers_Y += self.line_gap


def render_line_contents_by_dicts(self, dicts) -> None:
    # Preparation of the rendering:
    self.yline = self.yline_start
    first_line = self.showStartLine
    if self.showable_line_numbers_in_editor < len(self.line_string_list):
        # we got more text than we are able to display
        last_line = self.showStartLine + self.showable_line_numbers_in_editor
    else:
        last_line = self.maxLines

    # Actual line rendering based on dict-keys
    for line_list in dicts[first_line: last_line]:
        xcoord = self.xline_start
        for dict in line_list:
            surface = self.courier_font.render(dict['chars'], 1, dict['color'])  # create surface
            self.screen.blit(surface, (xcoord, self.yline))  # blit surface onto screen
            xcoord = xcoord + (len(dict['chars']) * self.letter_size_X)  # next line-part prep

        self.yline += self.line_gap  # next line prep


def caret_within_texteditor(self) -> bool:
    """
    Tests whether the caret's coordinates are within the visible text area.
    If the caret can possibly be in a line which is not currently displayed after using the mouse wheel for scrolling.
    Test for 'self.editor_offset_Y <= self.cursor_Y' as the caret can have the exact same Y-coordinate as the offset if
    the caret is in the first line.
    """
    return self.editor_offset_X + self.lineNumberWidth < self.cursor_X < (
                self.editor_offset_X + self.textAreaWidth - self.scrollBarWidth - 2) \
           and self.editor_offset_Y <= self.cursor_Y < (
                       self.textAreaHeight + self.editor_offset_Y - self.conclusionBarHeight)


def render_caret(self) -> None:
    """
    Called every frame. Displays a cursor for x frames, then none for x frames. Only displayed if line in which
    caret resides is visible and there is no active dragging operation going on.
    Dependent on FPS -> 5 intervalls per second
    Creates 'blinking' animation
    """
    self.Trenn_counter += 1
    if self.Trenn_counter > (self.FPS / 5) and self.caret_within_texteditor() and self.dragged_finished:
        self.screen.blit(self.trennzeichen_image, (self.cursor_X, self.cursor_Y))
        self.Trenn_counter = self.Trenn_counter % ((self.FPS / 5) * 2)


def reset_text_area_to_caret(self) -> None:
    """
    Reset visual area to include the line of caret if it is currently not visible. This function ensures
    that whenever we type, the line in which the caret resides becomes visible, even after scrolling.
    """
    if self.chosen_LineIndex < self.showStartLine:  # above visible area
        self.showStartLine = self.chosen_LineIndex
        self.rerenderLineNumbers = True
        self.update_caret_position()
    elif self.chosen_LineIndex > (self.showStartLine + self.showable_line_numbers_in_editor - 1):  # below visible area
        self.showStartLine = self.chosen_LineIndex - self.showable_line_numbers_in_editor + 1
        self.rerenderLineNumbers = True
        self.update_caret_position()
    # TODO: set cursor coordinates



import pygame
import pyperclip


def handle_input_with_highlight(self, input_event) -> None:
    """
    Handles key-downs after a drag operation was finished and the highlighted area (drag) is still active.
    For arrow keys we merely jump to the destination.
    For other character-keys we remove the highlighted area and replace it (also over multiple lines)
    with the chosen letter.
    """
    # for readability & maintainability we use shorter variable names
    line_start = self.drag_chosen_LineIndex_start
    line_end = self.drag_chosen_LineIndex_end
    letter_start = self.drag_chosen_LetterIndex_start
    letter_end = self.drag_chosen_LetterIndex_end

    if self.dragged_finished and self.dragged_active:
        if input_event.key in (pygame.K_DOWN, pygame.K_UP, pygame.K_RIGHT, pygame.K_LEFT):
            # deselect highlight
            if input_event.key == pygame.K_DOWN:
                self.jump_to_end(line_start, line_end, letter_start, letter_end)
            elif input_event.key == pygame.K_UP:
                self.jump_to_start(line_start, line_end, letter_start, letter_end)
            elif input_event.key == pygame.K_RIGHT:
                self.jump_to_end(line_start, line_end, letter_start, letter_end)
            elif input_event.key == pygame.K_LEFT:
                self.jump_to_start(line_start, line_end, letter_start, letter_end)
            self.reset_after_highlight()

        elif input_event.key in (pygame.K_RSHIFT, pygame.K_LSHIFT, pygame.K_CAPSLOCK, pygame.K_RCTRL, pygame.K_LCTRL):
            pass  # nothing happens, we wait for the second key with which it is being used in combination

        else:  # other key -> delete highlighted area and insert key (if not esc/delete)
            if line_start == line_end:  # delete in single line
                if letter_start > letter_end:  # swap variables based on left/right highlight to make code more readable
                    letter_start, letter_end = letter_end, letter_start
                self.delete_letter_to_letter(line_start, letter_start, letter_end)

            else:  # multi-line delete
                if line_start > line_end:  # swap variables based on up/downward highlight to make code more readable
                    line_start, line_end = line_end, line_start
                    letter_start, letter_end = letter_end, letter_start

                for i, line_number in enumerate(range(line_start, line_end + 1)):
                    if i == 0:  # first line
                        self.delete_letter_to_end(line_start, letter_start)  # delete right side from start
                    elif i < (line_end - line_start):
                        self.delete_entire_line(line_start + 1)  # stays at line_start +1 as we delete on the fly (!)
                    else:  # last line
                        self.delete_start_to_letter(line_start + 1, letter_end)  # delete left side of new last line

                # join rest of start/end lines into new line in multiline delete
                self.line_string_list[line_start] = self.line_string_list[line_start] + \
                                                     self.line_string_list[line_start + 1]
                self.delete_entire_line(line_start + 1)  # after copying contents, we need to delete the other line

            # set caret and rerender line_numbers
            self.chosen_LineIndex = line_start if line_start <= line_end else line_end  # start for single_line
            self.chosen_LetterIndex = letter_start if line_start <= line_end else letter_end
            self.rerenderLineNumbers = True
            self.reset_after_highlight()

            # insert key unless delete/backspace
            if input_event.key not in (pygame.K_DELETE, pygame.K_BACKSPACE):
                self.insert_unicode(input_event.unicode)


def handle_highlight_and_paste(self):
    """
    Paste clipboard into cursor position.
    Replace highlighted area if highlight, else normal insert.
    """

    # DELETE highlighted section if something is highlighted
    if self.dragged_finished and self.dragged_active:
        delete_event = pygame.event.Event(pygame.KEYDOWN, key=pygame.K_DELETE)  # create artifical delete event
        self.handle_input_with_highlight(delete_event)

    # PASTE from clipboard
    paste_string = pyperclip.paste()
    line_split = paste_string.split("\r\n")  # split into lines
    if len(line_split) == 1:  # no linebreaks
        self.line_string_list[self.chosen_LineIndex] = \
            self.line_string_list[self.chosen_LineIndex][:self.chosen_LetterIndex] \
            + line_split[0] + \
            self.line_string_list[self.chosen_LineIndex][self.chosen_LetterIndex:]

        self.chosen_LetterIndex = self.chosen_LetterIndex + len(line_split[0])

    else:
        rest_of_line = self.line_string_list[self.chosen_LineIndex][self.chosen_LetterIndex:]  # store for later
        for i, line in enumerate(line_split):
            if i == 0:  # first line to insert
                self.line_string_list[self.chosen_LineIndex] = \
                    self.line_string_list[self.chosen_LineIndex][:self.chosen_LetterIndex] + line
            elif i < len(line_split) - 1:  # middle line -> insert new line!
                self.line_string_list[self.chosen_LineIndex + i: self.chosen_LineIndex + i] = [line]
                self.maxLines += 1
            else:  # last line
                self.line_string_list[self.chosen_LineIndex + i: self.chosen_LineIndex + i] = [line + rest_of_line]
                self.maxLines += 1

                self.chosen_LetterIndex = len(line)
                self.chosen_LineIndex = self.chosen_LineIndex + i

    self.update_caret_position()
    self.rerenderLineNumbers = True


def handle_highlight_and_copy(self):
    """
    Copy highlighted String into clipboard if anything is highlighted, else no action.
    """
    copy_string = self.get_highlighted_characters()
    pyperclip.copy(copy_string)


def handle_highlight_and_cut(self):
    """
    Copy highlighted String into clipboard if anything is highlighted, else no action.
    Delete highlighted part of the text.
    """
    # Copy functionality
    copy_string = self.get_highlighted_characters()  # copy characters
    pyperclip.copy(copy_string)

    # Cut / delete functionality
    delete_event = pygame.event.Event(pygame.KEYDOWN, key=pygame.K_DELETE)  # create artifical event
    self.handle_input_with_highlight(delete_event)

    self.update_caret_position()

def highlight_all(self):
    """
    Highlight entire text.
    """
    # set artifical drag and cursor position
    self.set_drag_start_before_first_line()
    self.set_drag_end_after_last_line()
    self.update_caret_position_by_drag_end()
    # activate highlight
    self.dragged_finished = True
    self.dragged_active = True


def get_highlighted_characters(self) -> str:
    """
    Returns the highlighted characters (single- and multiple-line) from the editor (self.line_string_list)
    """
    if self.dragged_finished and self.dragged_active:
        line_start = self.drag_chosen_LineIndex_start
        line_end = self.drag_chosen_LineIndex_end
        letter_start = self.drag_chosen_LetterIndex_start
        letter_end = self.drag_chosen_LetterIndex_end

        if self.drag_chosen_LineIndex_start == self.drag_chosen_LineIndex_end:
            # single-line highlight
            return self.get_line_from_char_to_char(self.drag_chosen_LineIndex_start, self.drag_chosen_LetterIndex_start,
                                                   self.drag_chosen_LetterIndex_end)

        else:  # multi-line highlight
            if line_start > line_end:  # swap variables based on up/downward highlight to make code more readable
                line_start, line_end = line_end, line_start
                letter_start, letter_end = letter_end, letter_start

            # loop through highlighted lines
            copied_chars = ""
            for i, line_index in enumerate(range(line_start, line_end+1)):
                if i == 0:  # first line
                    copied_chars = self.get_line_from_char_to_end(line_index, letter_start)
                elif i < len(range(line_start, line_end)):  # middle line
                    copied_chars = copied_chars + "\r\n" + self.get_entire_line(line_index)
                else:  # last line
                    copied_chars = copied_chars + "\r\n" + self.get_line_from_start_to_char(line_index, letter_end)

            return copied_chars
    else:
        return ""




import os
import re
import pathlib
import yaml
import pygame
from .PybrainzzFormatter import PybrainzzFormatter
from pygments.lexers import PythonLexer


def set_line_numbers(self, b) -> None:
    """
    Activates/deactivates showing the line numbers in the editor
    """
    self.displayLineNumbers = b
    if self.displayLineNumbers:
        self.lineNumberWidth = 27  # line number background width and also offset for text!
        self.xline_start = self.editor_offset_X + self.xline_start_offset
        self.xline = self.editor_offset_X + self.xline_start_offset
    else:
        self.lineNumberWidth = 0
        self.xline_start = self.editor_offset_X
        self.xline = self.editor_offset_X


def set_syntax_highlighting(self, b) -> None:
    """
    Activates / deactivates syntax highlighting.
    If activated, creates the lexer and formatter and sets the formatter's style.
    """
    self.syntax_coloring = b


def set_colorscheme_from_yaml(self, path_to_yaml) -> None:

    def get_rgb_by_key(dict, key):
        return tuple([int(x) for x in re.findall(r'(\d{1,3})', dict[key])])

    try:
        with open(path_to_yaml) as file:
            color_dict = yaml.load(file, Loader=yaml.FullLoader)

            self.codingBackgroundColor = get_rgb_by_key(color_dict, 'codingBackgroundColor')
            self.codingScrollBarBackgroundColor = get_rgb_by_key(color_dict, 'codingScrollBarBackgroundColor')
            self.lineNumberColor = get_rgb_by_key(color_dict, 'lineNumberColor')
            self.lineNumberBackgroundColor = get_rgb_by_key(color_dict, 'lineNumberBackgroundColor')
            self.textColor = get_rgb_by_key(color_dict, 'textColor')

            self.formatter.textColor_normal = get_rgb_by_key(color_dict, 'textColor_normal')
            self.formatter.textColor_comments = get_rgb_by_key(color_dict, 'textColor_comments')
            self.formatter.textColor_quotes = get_rgb_by_key(color_dict, 'textColor_quotes')
            self.formatter.textColor_operators = get_rgb_by_key(color_dict, 'textColor_operators')
            self.formatter.textColor_keywords = get_rgb_by_key(color_dict, 'textColor_keywords')
            self.formatter.textColor_function = get_rgb_by_key(color_dict, 'textColor_function')
            self.formatter.textColor_builtin = get_rgb_by_key(color_dict, 'textColor_builtin')

    except FileNotFoundError:
        raise FileNotFoundError("Could not find the style file '" + path_to_yaml + "'.")
    except KeyError as e:
        raise KeyError("Could not find all necessary color-keys in the style file '" + path_to_yaml + "'. " +
                       "Missing key: " + str(e))


def set_colorscheme(self, style) -> None:
    """
    Sets the color scheme for the editor and for syntax highlighting if the latter is enabled based on
    one of the editor's default styles.
    """
    if style in ("bright", "dark"):
        default_path = str(pathlib.Path(__file__).parent.absolute()) + "\\elements\colorstyles\\" + style + ".yml"
        self.set_colorscheme_from_yaml(default_path)
    else:
        raise ValueError("No default style with the name '" + style + "' available. " +
                         "\n\tAvailable styles are 'dark' and 'bright'. \n\tFor your " +
                         "own custom styles, use the method 'set_colorscheme_from_yaml(path_to_yaml)' " +
                         "and have a look at the docs.")





def set_drag_start_by_mouse(self, mouse_x, mouse_y) -> None:
    # get line
    self.drag_chosen_LineIndex_start = self.get_line_index(mouse_y)

    # end of line
    if self.get_number_of_letters_in_line_by_mouse(mouse_y) < self.get_letter_index(mouse_x):
        self.drag_chosen_LetterIndex_start = len(self.line_string_list[self.drag_chosen_LineIndex_start])

    else:  # within existing line
        self.drag_chosen_LetterIndex_start = self.get_letter_index(mouse_x)


def set_drag_end_by_mouse(self, mouse_x, mouse_y) -> None:
    """
    Compact method to set both line and letter of drag_end based on mouse coordinates.
    """
    # set line
    self.set_drag_end_line_by_mouse(mouse_y)
    # set letter
    self.set_drag_end_letter_by_mouse(mouse_x)


def set_drag_end_line_by_mouse(self, mouse_y) -> None:
    """
    Sets self.drag_chosen_LineIndex_end by mouse_y.
    """
    self.drag_chosen_LineIndex_end = self.get_line_index(mouse_y)


def set_drag_end_letter_by_mouse(self, mouse_x) -> None:
    """
    Sets self.drag_chosen_LetterIndex_end by mouse_x.
    Dependent on self.drag_chosen_LineIndex_end.
    """
    # end of line
    if self.get_letter_index(mouse_x) > self.get_number_of_letters_in_line_by_index(self.drag_chosen_LineIndex_end):
        self.drag_chosen_LetterIndex_end = len(self.line_string_list[self.drag_chosen_LineIndex_end])
    else:  # within existing line
        self.drag_chosen_LetterIndex_end = self.get_letter_index(mouse_x)


def set_drag_start_after_last_line(self) -> None:
    # select last line
    self.drag_chosen_LineIndex_start = self.maxLines - 1
    # select last letter of the line
    self.drag_chosen_LetterIndex_start = len(self.line_string_list[self.drag_chosen_LineIndex_start])


def set_drag_start_before_first_line(self) -> None:
    self.drag_chosen_LineIndex_start = 0
    self.drag_chosen_LetterIndex_start = 0


def set_drag_end_after_last_line(self) -> None:
    # select last line
    self.drag_chosen_LineIndex_end = self.maxLines - 1
    # select last letter of the line
    self.drag_chosen_LetterIndex_end = len(self.line_string_list[self.drag_chosen_LineIndex_end])


def update_caret_position_by_drag_start(self) -> None:
    """
    # Updates cursor_X and cursor_Y positions based on the start position of a dragging operation.
    """
    # X Position
    self.cursor_X = self.xline_start + (self.drag_chosen_LetterIndex_start * self.letter_size_X)
    # Y Position
    self.cursor_Y = self.editor_offset_Y + (self.drag_chosen_LineIndex_start * self.line_gap) - \
                    (self.showStartLine * self.lineHeight)


def update_caret_position_by_drag_end(self) -> None:
    """
    # Updates cursor_X and cursor_Y positions based on the end position of a dragging operation.
    """
    # X Position
    self.cursor_X = self.xline_start + (self.drag_chosen_LetterIndex_end * self.letter_size_X)
    # Y Position
    self.cursor_Y = self.editor_offset_Y + (self.drag_chosen_LineIndex_end * self.line_gap) - \
                    (self.showStartLine * self.lineHeight)


def update_caret_position(self) -> None:
    """
    # Updates cursor_X and cursor_Y positions based on current position by line and letter indices
    """
    self.cursor_X = self.xline_start + (self.chosen_LetterIndex * self.letter_size_X)

    self.cursor_Y = self.editor_offset_Y + (self.chosen_LineIndex * self.line_gap) - \
                    (self.showStartLine * self.lineHeight)



import pygame
import math
from typing import List


def get_text_as_string(self) -> str:
    """
    Returns the entire text of the editor as a single string.
    Linebreak characters are used to differentiate between lines.
    :param self:  Texteditor-Class
    :return: String
    """
    return "\n".join(self.line_string_list)


def get_text_as_list(self) -> List:
    """
    Returns the text in it's logical form as a list of lines.
    :param self:  Texteditor-Class
    :return: List of lines containing the text. Lines cane be empty Strings.
    """
    return self.line_string_list


def clear_text(self) -> None:
    """
    Clears the textarea.
    :param self: Texteditor-Class
    :return: None
    """
    self.line_string_list = []  # LOGIC: List of actual Strings for each line
    self.maxLines = int(math.floor(self.textAreaHeight / self.lineHeight))
    self.showStartLine = 0
    for i in range(self.maxLines):  # from 0 to maxLines:
        self.line_string_list.append("")  # Add a line

    # reset caret
    self.firstiteration_boolean = True  # redraws background
    self.rerenderLineNumbers = True
    self.chosen_LineIndex = 0
    self.chosen_LetterIndex = 0
    self.dragged_active = False
    self.dragged_finished = True
    self.scroll_dragging = False
    self.drag_chosen_LineIndex_start = 0
    self.drag_chosen_LetterIndex_start = 0
    self.drag_chosen_LineIndex_end = 0
    self.drag_chosen_LetterIndex_end = 0
    self.last_clickdown_cycle = 0
    self.last_clickup_cycle = 0
    self.cycleCounter = 0
    self.update_caret_position()
    self.cycleCounter = 0


def set_text_from_list(self, text_list) -> None:
    """
    Sets the text of the editor based on a list of strings. Each item in the list represents one line.
    """
    self.clear_text()
    self.line_string_list = text_list
    self.maxLines = len(self.line_string_list)
    self.rerenderLineNumbers = True


def set_text_from_string(self, string) -> None:
    """
    Sets the text of the editor based on a string. Linebreak characters are parsed.
    """
    self.clear_text()
    self.line_string_list = string.split('\n')
    self.maxLines = len(self.line_string_list)
    self.rerenderLineNumbers = True



import pygame


def handle_keyboard_input(self, pygame_events, pressed_keys) -> None:

    if self.deleteCounter > 0:
        self.deleteCounter -= 1

    # Detect tapping/holding of the "DELETE" and "BACKSPACE" key
    if self.dragged_finished and self.dragged_active and \
            (pressed_keys[pygame.K_DELETE] or pressed_keys[pygame.K_BACKSPACE]):
        delete_event = pygame.event.Event(pygame.KEYDOWN, key=pygame.K_DELETE)  # create the event
        self.handle_input_with_highlight(delete_event)  # delete and backspace have the same functionality
        self.deleteCounter = 5
    elif pressed_keys[pygame.K_DELETE] and self.deleteCounter == 0:
        self.handle_keyboard_delete()  # handle input
        self.reset_text_area_to_caret()  # reset caret if necessary
        self.deleteCounter = 5
    elif pressed_keys[pygame.K_BACKSPACE] and self.deleteCounter == 0:
        self.handle_keyboard_backspace()  # handle input
        self.reset_text_area_to_caret()  # reset caret if necessary
        self.deleteCounter = 5

    # ___ OTHER KEYS ___ #
    for event in pygame_events:
        if event.type == pygame.KEYDOWN:

            # ___ COMBINATION KEYS ___
            # Can be applied whether something is highlighted or not!
            if (pressed_keys[pygame.K_LCTRL] or pressed_keys[pygame.K_RCTRL]) and event.key == pygame.K_a:
                self.highlight_all()
            elif (pressed_keys[pygame.K_LCTRL] or pressed_keys[pygame.K_RCTRL]) and event.key == pygame.K_v:
                self.handle_highlight_and_paste()

            elif self.dragged_finished and self.dragged_active:
                # Only work if something is highlighted!
                if (pressed_keys[pygame.K_LCTRL] or pressed_keys[pygame.K_RCTRL]) and event.key == pygame.K_x:
                    self.handle_highlight_and_cut()
                elif (pressed_keys[pygame.K_LCTRL] or pressed_keys[pygame.K_RCTRL]) and event.key == pygame.K_c:
                    self.handle_highlight_and_copy()
                else:
                    self.handle_input_with_highlight(event)  # handle char input on highlight

            else:
                self.reset_text_area_to_caret()  # reset visual area to include line of caret if necessaryss

                # ___ NORMAL KEYS ___
                if len(pygame.key.name(event.key)) == 1:  # This covers all letters and numbers (not on numpad).
                    self.chosen_LetterIndex = int(self.chosen_LetterIndex)
                    self.insert_unicode(event.unicode)
                # ___ SPECIAL KEYS ___
                elif event.key == pygame.K_TAB:  # ___TABULATOR
                    self.handle_keyboard_tab()
                elif event.key == pygame.K_SPACE:  # ___SPACEBAR
                    self.handle_keyboard_space()
                elif event.key == pygame.K_RETURN:  # ___RETURN
                    self.handle_keyboard_return()
                elif event.key == pygame.K_UP:  # ___ARROW_UP
                    self.handle_keyboard_arrow_up()
                elif event.key == pygame.K_DOWN:  # ___ARROW_DOWN
                    self.handle_keyboard_arrow_down()
                elif event.key == pygame.K_RIGHT:  # ___ARROW_RIGHT
                    self.handle_keyboard_arrow_right()
                elif event.key == pygame.K_LEFT:  # ___ARROW_LEFT
                    self.handle_keyboard_arrow_left()


                else:
                    if event.key not in [pygame.K_RSHIFT, pygame.K_LSHIFT, pygame.K_DELETE,
                                         pygame.K_BACKSPACE, pygame.K_CAPSLOCK, pygame.K_LCTRL, pygame.K_RCTRL]:
                        # We handled the keys separately
                        # Capslock is apparently implicitly handled when using it in combination

                        # disabled for smooth development:
                        # raise ValueError("No key implementation: " + str(pygame.key.name(event.key)))
                        print("No key implementation: " + str(pygame.key.name(event.key)))


def insert_unicode(self, unicode) -> None:
    self.line_string_list[self.chosen_LineIndex] = self.line_string_list[self.chosen_LineIndex][
                                                    :self.chosen_LetterIndex] + unicode + \
                                                    self.line_string_list[self.chosen_LineIndex][
                                                    self.chosen_LetterIndex:]
    self.cursor_X += self.letter_size_X
    self.chosen_LetterIndex += 1


def handle_keyboard_backspace(self) -> None:
    if self.chosen_LetterIndex == 0 and self.chosen_LineIndex == 0:
        # First position and in the first Line -> nothing happens
        pass
    elif self.chosen_LetterIndex == 0 and self.chosen_LineIndex > 0:
        # One Line back if at X-Position 0 and not in the first Line

        # set letter and line index to newly current line
        self.chosen_LineIndex -= 1
        self.chosen_LetterIndex = len(self.line_string_list[self.chosen_LineIndex])
        # set visual cursor one line above and at the end of the line
        self.cursor_Y -= self.line_gap
        self.cursor_X = self.xline_start + (len(self.line_string_list[self.chosen_LineIndex]) * self.letter_size_X)

        # take the rest of the former line into the current line
        self.line_string_list[self.chosen_LineIndex] = self.line_string_list[self.chosen_LineIndex] + \
            self.line_string_list[self.chosen_LineIndex + 1]

        # delete the former line
        # LOGICAL lines
        self.line_string_list.pop(self.chosen_LineIndex + 1)
        self.maxLines -= 1
        # VISUAL lines
        self.rerenderLineNumbers = True

        # Handling of the resulting scrolling functionality of removing one line
        if self.showStartLine > 0:
            if (self.showStartLine + self.showable_line_numbers_in_editor) > self.maxLines:
                # The scrollbar is all the way down. We delete a line,
                # so we have to "pull everything one visual line down"
                self.showStartLine -= 1  # "pull one visual line down" (array-based)
                self.cursor_Y += self.line_gap  # move the curser one down.  (visually based)
        if self.chosen_LineIndex == (self.showStartLine - 1):
            # Im in the first rendered line (but NOT  the "0" line) and at the beginning of the line.
            # => move one upward, change showstartLine & cursor placement.
            self.showStartLine -= 1
            self.cursor_Y += self.line_gap

    elif self.chosen_LetterIndex > 0:
        # mid-line or end of the line -> Delete a letter
        self.line_string_list[self.chosen_LineIndex] = self.line_string_list[self.chosen_LineIndex]\
            [:(self.chosen_LetterIndex - 1)] + self.line_string_list[self.chosen_LineIndex][self.chosen_LetterIndex:]

        self.cursor_X -= self.letter_size_X
        self.chosen_LetterIndex -= 1
    else:
        raise ValueError("INVALID CONSTRUCT: handle_keyboard_backspace. \
                         \nLine:" + str(self.chosen_LineIndex) + "\nLetter: " + str(self.chosen_LetterIndex))


def handle_keyboard_delete(self) -> None:

    if self.chosen_LetterIndex < (len(self.line_string_list[self.chosen_LineIndex])):
        # start of the line or mid-line (Cursor stays on point), cut one letter out
        self.line_string_list[self.chosen_LineIndex] = self.line_string_list[self.chosen_LineIndex] \
             [:self.chosen_LetterIndex] + self.line_string_list[self.chosen_LineIndex][(self.chosen_LetterIndex + 1):]

    elif self.chosen_LetterIndex == len(self.line_string_list[self.chosen_LineIndex]):
        # End of a line  (choose next line)
        if self.chosen_LineIndex != (
                self.maxLines - 1):  # NOT in the last line &(prev) at the end of the line, I cannot delete anything
            self.line_string_list[self.chosen_LineIndex] += self.line_string_list[
                self.chosen_LineIndex + 1]  # add the contents of the next line to the current one
            self.line_string_list.pop(
                self.chosen_LineIndex + 1)  # delete the Strings-line in order to move the following lines one upwards
            self.maxLines -= 1  # Keep the variable aligned
            self.rerenderLineNumbers = True
            if self.showStartLine > 0:
                if (self.showStartLine + self.showable_line_numbers_in_editor) > self.maxLines:
                    # The scrollbar is all the way down.
                    # We delete a line, so we have to "pull everything one visual line down"
                    self.showStartLine -= 1  # "pull one visual line down" (array-based)
                    self.cursor_Y += self.line_gap  # move the curser one down.  (visually based)
    else:
        raise ValueError(" INVALID CONSTRUCT: handle_keyboard_delete. \
                         \nLine:" + str(self.chosen_LineIndex) + "\nLetter: " + str(self.chosen_LetterIndex))


def handle_keyboard_arrow_left(self) -> None:
    if self.chosen_LetterIndex > 0:  # mid-line or end of line
        self.chosen_LetterIndex -= 1
        self.cursor_X -= self.letter_size_X
    elif self.chosen_LetterIndex == 0 and self.chosen_LineIndex == 0:
        # first line, first position, nothing happens
        pass
    elif self.chosen_LetterIndex == 0 and self.chosen_LineIndex > 0:  # Move over into previous Line (if there is any)
        self.chosen_LineIndex -= 1
        self.chosen_LetterIndex = len(self.line_string_list[self.chosen_LineIndex])  # end of previous line
        self.cursor_X = self.xline_start + (len(self.line_string_list[self.chosen_LineIndex]) * self.letter_size_X)
        self.cursor_Y -= self.line_gap
        if self.chosen_LineIndex < self.showStartLine:
            # handling scroll functionality if necessary (moved above shown lines)
            self.showStartLine -= 1
            self.cursor_Y += self.line_gap
            self.rerenderLineNumbers = True


def handle_keyboard_arrow_right(self) -> None:
    if self.chosen_LetterIndex < (len(self.line_string_list[self.chosen_LineIndex])):
        # mid-line or start of the line
        self.chosen_LetterIndex += 1
        self.cursor_X += self.letter_size_X
    elif self.chosen_LetterIndex == len(self.line_string_list[self.chosen_LineIndex]) and \
            not (self.chosen_LineIndex == (self.maxLines - 1)):
        # end of line => move over into the start of the next line

        self.chosen_LetterIndex = 0
        self.chosen_LineIndex += 1
        self.cursor_X = self.xline_start
        self.cursor_Y += self.line_gap
        if self.chosen_LineIndex > (self.showStartLine + self.showable_line_numbers_in_editor - 1):
            # handling scroll functionality if necessary (moved below showed lines)
            self.showStartLine += 1
            self.cursor_Y -= self.line_gap
            self.rerenderLineNumbers = True


def handle_keyboard_arrow_down(self) -> None:
    if self.chosen_LineIndex < (self.maxLines - 1):
        # Not in the last line, downward movement possible
        self.chosen_LineIndex += 1
        self.cursor_Y += self.line_gap

        if len(self.line_string_list[self.chosen_LineIndex]) < self.chosen_LetterIndex:
            # reset letter-index to the end
            self.chosen_LetterIndex = len(self.line_string_list[self.chosen_LineIndex])
            self.cursor_X = (len(self.line_string_list[self.chosen_LineIndex]) * self.letter_size_X) + self.xline_start

        if self.chosen_LineIndex > (self.showStartLine + self.showable_line_numbers_in_editor - 1):
            # handle scrolling functionality if necessary (moved below shown lines)
            self.scrollDown()

    elif self.chosen_LineIndex == (self.maxLines - 1):  # im in the last line and want to jump to its end.
        self.chosen_LetterIndex = len(self.line_string_list[self.chosen_LineIndex])  # end of the line
        self.cursor_X = self.xline_start + (
                len(self.line_string_list[self.chosen_LineIndex]) * self.letter_size_X)


def handle_keyboard_arrow_up(self) -> None:
    if self.chosen_LineIndex == 0:
        # first line, cannot go upwards, so we go to the first position
        self.chosen_LetterIndex = 0
        self.cursor_X = self.xline_start

    elif self.chosen_LineIndex > 0:
        # subsequent lines, upwards movement possible
        self.chosen_LineIndex -= 1
        self.cursor_Y -= self.line_gap

        if len(self.line_string_list[self.chosen_LineIndex]) < self.chosen_LetterIndex:
            # less letters in this line, reset toward the end of the line (to the left)
            self.chosen_LetterIndex = len(self.line_string_list[self.chosen_LineIndex])
            self.cursor_X = (len(self.line_string_list[self.chosen_LineIndex])) * self.letter_size_X + self.xline_start

        if self.chosen_LineIndex < self.showStartLine:  # scroll up one line
            self.scrollUp()


def handle_keyboard_tab(self) -> None:
    for x in range(0, 4):  # insert 4 spaces
        self.handle_keyboard_space()


def handle_keyboard_space(self) -> None:
    # insert 1 space
    self.line_string_list[self.chosen_LineIndex] = \
        self.line_string_list[self.chosen_LineIndex][:self.chosen_LetterIndex] + " " + \
        self.line_string_list[self.chosen_LineIndex][self.chosen_LetterIndex:]

    self.cursor_X += self.letter_size_X
    self.chosen_LetterIndex += 1


def handle_keyboard_return(self) -> None:
    # Get "transfer letters" behind cursor up to the end of the line to next line
    # If the cursor is at the end of the line, transferString is an empty String ("")
    transferString = self.line_string_list[self.chosen_LineIndex][self.chosen_LetterIndex:]

    # Remove transfer letters from the current line
    self.line_string_list[self.chosen_LineIndex] = \
        self.line_string_list[self.chosen_LineIndex][:self.chosen_LetterIndex]

    # set logical cursor indizes and add a new line
    self.chosen_LineIndex += 1
    self.chosen_LetterIndex = 0
    self.maxLines += 1
    self.cursor_X = self.xline_start  # reset cursor to start of the line

    # insert empty line
    self.line_string_list.insert(self.chosen_LineIndex, "")  # logical line

    # Edit the new line -> append transfer letters
    self.line_string_list[self.chosen_LineIndex] = self.line_string_list[self.chosen_LineIndex] + transferString
    self.rerenderLineNumbers = True

    # handle scrolling functionality
    if self.chosen_LineIndex > (self.showable_line_numbers_in_editor - 1):  # Last row, visual representation moves down
        self.showStartLine += 1
    else:  # not in last row, put courser one line down without changing the shown line numbers
        self.cursor_Y += self.line_gap






def get_line_index(self, mouse_y) -> int:
    """
    Returns possible line-position of mouse -> does not take into account
    how many lines there actually are!
    """
    return int(((mouse_y - self.editor_offset_Y) / self.line_gap) + (self.showStartLine))


def get_letter_index(self, mouse_x) -> int:
    """
    Returns possible letter-position of mouse.

    The function is independent from any specific line, so we could possible return a letter_index which
    is bigger than the letters in the line.
    Returns at least 0 to make sure it is possibly a valid index.
    """
    letter = int((mouse_x - self.xline_start) / self.letter_size_X)
    letter = 0 if letter < 0 else letter
    return letter


def get_number_of_letters_in_line_by_mouse(self, mouse_y) -> int:
    line_index = get_line_index(self, mouse_y)
    return get_number_of_letters_in_line_by_index(self, line_index)


def get_number_of_letters_in_line_by_index(self, index) -> int:
    return len(self.line_string_list[index])


def get_showable_lines(self) -> int:
    """
    Return the number of lines which are shown. Less than maximum if less lines are in the array.
    """
    if self.showable_line_numbers_in_editor + self.showStartLine < self.maxLines:
        return self.showable_line_numbers_in_editor + self.showStartLine
    else:
        return self.maxLines


def line_is_visible(self, line) -> bool:
    """
    Calculate whether the line is being shown in the editor
    """
    return self.showStartLine <= line < self.showStartLine + self.showable_line_numbers_in_editor




import pygame


def render_scrollbar_vertical(self) -> None:
    self.display_scrollbar()


def display_scrollbar(self):
    # Scroll Bar

    if len(self.line_string_list) <= self.showable_line_numbers_in_editor:
        self.scrollbar = None
        return  # if scrollbar is not needed, don't show.

    # scroll bar is a fraction of the space
    w = self.scrollBarWidth 
    x = self.editor_offset_X + self.textAreaWidth - self.scrollBarWidth - 2  # -2 for space between edge & scrollbar
    y = int(self.editor_offset_Y + (w/2) + (self.textAreaHeight * ((self.showStartLine * 1.0) / self.maxLines)))
    h = int((self.textAreaHeight - w) * ((self.showable_line_numbers_in_editor * 1.0) / self.maxLines))
    self.scrollbar = pygame.Rect(x, y, w, h)

    pygame.draw.circle(self.screen, self.color_scrollbar, (int(x + (w/2)), y), int(w/2))  # top round corner
    pygame.draw.rect(self.screen, self.color_scrollbar, self.scrollbar)  # actual scrollbar
    pygame.draw.circle(self.screen, self.color_scrollbar, (int(x + (w/2)), y + h), int(w/2))  # bottom round corner


def scrollbar_up(self) -> None:
    self.showStartLine -= 1
    self.cursor_Y += self.line_gap
    self.rerenderLineNumbers = True
    


def scrollbar_down(self) -> None:
    self.showStartLine += 1
    self.cursor_Y -= self.line_gap
    self.rerenderLineNumbers = True






def jump_to_start(self, line_start, line_end, letter_start, letter_end) -> None:
    """
    Chosen LineIndex set to start of highlighted area
    """
    if line_start <= line_end:
        # downward highlight or the same line
        self.chosen_LineIndex = line_start
        self.chosen_LetterIndex = letter_start
    else:  # upward highlight
        self.chosen_LineIndex = line_end
        self.chosen_LetterIndex = letter_end


def jump_to_end(self, line_start, line_end, letter_start, letter_end) -> None:
    """
    Chosen LineIndex set to end of highlighted area
    """
    if line_start <= line_end:
        # downward highlight or the same line
        self.chosen_LineIndex = line_end
        self.chosen_LetterIndex = letter_end
    else:  # upward highlight
        self.chosen_LineIndex = line_start
        self.chosen_LetterIndex = letter_start


def reset_after_highlight(self) -> None:
    """
    Reset caret, clickdown_cycles and dragged booleans.
    """
    self.dragged_active = False  # deactivate highlight
    self.dragged_finished = True  # highlight is finished
    self.update_caret_position()  # update caret position to chosen_Index (Line+Letter)
    self.last_clickdown_cycle = 0  # reset drag-cycle
    self.last_clickup_cycle = -1
    self.rerenderLineNumbers = True

    if len(self.line_string_list) <= self.showable_line_numbers_in_editor:
        self.showStartLine = 0  # update first showable line



import pygame
from ._scrollbar_vertical import scrollbar_up, scrollbar_up


def handle_mouse_input(self, pygame_events, mouse_x, mouse_y, mouse_pressed) -> None:
    """
    Handles mouse input based on mouse events (Buttons down/up + coordinates).
    Handles drag-and-drop-select as well as single-click.
    The code only differentiates the single-click only as so far, that
        the DOWN-event is on the same position as the UP-event.

    Implemented so far:
    - left-click (selecting as drag-and-drop or single-click)
    - mouse-wheel (scrolling)
    TODO:
    - right-click
    """

    # ___ MOUSE ___ #
    for event in pygame_events:
        # ___ MOUSE CLICKING DOWN ___ #
        # Scrollbar-handling
        if event.type == pygame.MOUSEBUTTONDOWN and not self.mouse_within_texteditor(mouse_x, mouse_y):
            if self.scrollbar is not None:
                if self.scrollbar.collidepoint(mouse_x, mouse_y):
                    self.scroll_start_y = mouse_y
                    self.scroll_dragging = True

        # Mouse scrolling wheel should only work if it is within the coding area (excluding scrollbar area)
        if event.type == pygame.MOUSEBUTTONDOWN and self.mouse_within_texteditor(mouse_x, mouse_y):
            # ___ MOUSE SCROLLING ___ #
            if event.button == 4 and self.showStartLine > 0:
                self.scrollbar_up()
            elif event.button == 5 and self.showStartLine + self.showable_line_numbers_in_editor < self.maxLines:
                self.scrollbar_down()

            # ___ MOUSE LEFT CLICK DOWN ___ #
            elif event.button == 1:  # left mouse button
                if not self.click_hold:
                    # in order not to have the mouse move around after a click,
                    # we need to disable this function until we RELEASE it.
                    self.last_clickdown_cycle = self.cycleCounter
                    self.click_hold = True
                    self.dragged_active = True
                    self.dragged_finished = False
                    if self.mouse_within_texteditor(mouse_x, mouse_y):  # editor area
                        if self.mouse_within_existing_lines(mouse_y):  # in area of existing lines
                            self.set_drag_start_by_mouse(mouse_x, mouse_y)
                        else:  # clicked below the existing lines
                            self.set_drag_start_after_last_line()
                        self.update_caret_position_by_drag_start()
                    else:  # mouse outside of editor, don't care.
                        pass

        # ___ MOUSE LEFT CLICK UP ___ #
        if event.type == pygame.MOUSEBUTTONUP and event.button == 1:

            self.scroll_dragging = False  # reset scroll (if necessary)

            if self.click_hold:
                # mouse dragging only with left mouse up
                # mouse-up only valid if we registered a mouse-down within the editor via click_hold earlier

                self.last_clickup_cycle = self.cycleCounter
                self.click_hold = False

                if self.mouse_within_texteditor(mouse_x, mouse_y):  # editor area
                    if self.mouse_within_existing_lines(mouse_y):  # in area of existing lines
                        self.set_drag_end_by_mouse(mouse_x, mouse_y)
                    else:  # clicked beneath the existing lines
                        self.set_drag_end_after_last_line()
                    self.update_caret_position_by_drag_end()

                else:  # mouse-up outside of editor
                    if mouse_y < self.editor_offset_Y:
                        # Mouse-up above editor -> set to first visible line
                        self.drag_chosen_LineIndex_end = self.showStartLine
                    elif mouse_y > (self.editor_offset_Y + self.textAreaHeight - self.conclusionBarHeight):
                        # Mouse-up below the editor -> set to last visible line
                        if self.maxLines >= self.showable_line_numbers_in_editor:
                            self.drag_chosen_LineIndex_end = self.showStartLine + self.showable_line_numbers_in_editor - 1
                        else:
                            self.drag_chosen_LineIndex_end = self.maxLines - 1
                    else:  # mouse left or right of the editor outside
                        self.set_drag_end_line_by_mouse(mouse_y)
                    # Now we can determine the letter based on mouse_x (and selected line within the function)
                    self.set_drag_end_letter_by_mouse(mouse_x)

        # _______ CHECK FOR MOUSE DRAG AND HANDLE CLICK _______ #
        if (self.last_clickup_cycle - self.last_clickdown_cycle) >= 0:
            # Clicked the mouse lately and has not been handled yet.
            # To differentiate between click and drag we check whether the down-click
            # is on the same letter and line as the up-click!
            # We set the boolean variables here and handle the caret positioning.
            if self.drag_chosen_LineIndex_end == self.drag_chosen_LineIndex_start and \
                    self.drag_chosen_LetterIndex_end == self.drag_chosen_LetterIndex_start:
                self.dragged_active = False  # no letters are actually selected -> Actual click
            else:
                self.dragged_active = True  # Actual highlight

            self.dragged_finished = True  # we finished the highlighting operation either way

            # handle caret positioning
            self.chosen_LineIndex = self.drag_chosen_LineIndex_end
            self.chosen_LetterIndex = self.drag_chosen_LetterIndex_end
            self.update_caret_position()

            # reset after upclick so we don't execute this block again and again.
            self.last_clickdown_cycle = 0
            self.last_clickup_cycle = -1

    # Scrollbar - Dragging
    if mouse_pressed[0] == 1 and self.scroll_dragging:
        # left mouse is being pressed after click on scrollbar
        if mouse_y < self.scroll_start_y and self.showStartLine > 0:
            # dragged higher
            self.scrollbar_up()
        elif mouse_y > self.scroll_start_y \
                and self.showStartLine + self.showable_line_numbers_in_editor < self.maxLines:
            # dragged lower
            self.scrollbar_down()

def mouse_within_texteditor(self, mouse_x, mouse_y) -> bool:
    """
    Returns True if the given coordinates are within the text-editor area of the pygame window, otherwise False.
    """
    return self.editor_offset_X + self.lineNumberWidth < mouse_x < (self.editor_offset_X + self.textAreaWidth - self.scrollBarWidth) \
           and self.editor_offset_Y < mouse_y < (self.textAreaHeight + self.editor_offset_Y - self.conclusionBarHeight)


def mouse_within_existing_lines(self, mouse_y):
    """
    Returns True if the given Y-coordinate is within the height of the text-editor's existing lines.
    Returns False if the coordinate is below existing lines or outside of the editor.
    """
    return self.editor_offset_Y < mouse_y < self.editor_offset_Y + (self.lineHeight * self.maxLines)




from typing import List, Dict, Tuple
import re as re
from pygments.lexers import PythonLexer
from .PybrainzzFormatter import PybrainzzFormatter


def get_single_color_dicts(self) -> List[List[Dict]]:
    """
    Converts the text in the editor based on the line_string_list into a list of lists of dicts.
    Every line is one sublist.
    Since only one color is being applied, we create a list with one dict per line.
    """
    rendering_list = []
    for line in self.line_string_list:
        # appends a single-item list
        rendering_list.append([{'chars': line, 'type': 'normal', 'color': self.textColor}])
    return rendering_list


def get_syntax_coloring_dicts(self) -> List[List[Dict]]:
    """
    Converts the text in the editor based on the line_string_list into a list of lists of dicts.
    Every line is one sublist which contains different dicts based on it's contents.

    We create a dict for every token of every line and include the characters, a pseudo token-type and the color.
    """
    rendering_list = []
    for line in self.line_string_list:
        line_dict = self.formatter.format(self.lexer.get_tokens(line))
        rendering_list.append(line_dict)
    return rendering_list



from setuptools import setup, find_packages
import pathlib

# The directory containing this file
HERE = pathlib.Path(__file__).parent

# The text of the README file
README = (HERE / "README.md").read_text()

setup(
    name='pygame-texteditor',
    packages=['pygame_texteditor'],
    package_data={'pygame_texteditor': ['pygame_texteditor/elements/graphics/*.png', 'pygame_texteditor/elements/fonts/.ttf', 'pygame_texteditor/elements/colorstyles/*.yml']},
    include_package_data=True,
    license='MIT',
    description='A WYSIWYG-texteditor based on pygame.',
    long_description=README,
    long_description_content_type='text/markdown',
    version='0.6.2',
    python_requires=">=3.6",
    author='CribberSix',
    author_email='cribbersix@gmail.com',
    install_requires=['pygame==1.9.6', 'pyyaml >= 5.3.1', 'pygments >= 2.6.1', 'pyperclip >= 1.8.1'],
    classifiers=[
        'Development Status :: 4 - Beta',
        # 'Development Status :: 5 - Production/Stable',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.6',
        'Topic :: Software Development :: Libraries',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'Topic :: Software Development',
    ],
    keywords=['pygame', 'texteditor', 'text', 'editor'],
)




from pygments.lexers import PythonLexer
from .PybrainzzFormatter import PybrainzzFormatter
import pygame
import math
import os
import yaml
import re


class TextEditor:
    # Scroll functionality
    from ._scrollbar_vertical import render_scrollbar_vertical, display_scrollbar, scrollbar_up, \
        scrollbar_down

    # input handling KEYBOARD
    from ._input_handling_keyboard import handle_keyboard_input, handle_keyboard_delete, handle_keyboard_backspace, \
        handle_keyboard_return, handle_keyboard_space, handle_keyboard_tab, insert_unicode, \
        handle_keyboard_arrow_left, handle_keyboard_arrow_right, handle_keyboard_arrow_up, handle_keyboard_arrow_down

    from ._input_handling_keyboard_highlight import handle_input_with_highlight, handle_highlight_and_copy, \
        handle_highlight_and_paste, handle_highlight_and_cut, highlight_all, get_highlighted_characters

    # input handling MOUSE
    from ._input_handling_mouse import handle_mouse_input, mouse_within_texteditor, mouse_within_existing_lines

    # caret
    from ._caret import update_caret_position, update_caret_position_by_drag_end, update_caret_position_by_drag_start, \
        set_drag_start_after_last_line, set_drag_end_after_last_line, set_drag_start_by_mouse, \
        set_drag_end_by_mouse, set_drag_end_line_by_mouse, set_drag_end_letter_by_mouse, \
        set_drag_start_before_first_line

    # letter operations (delete, get)
    from ._letter_operations import \
        delete_entire_line, delete_start_to_letter, delete_letter_to_end, delete_letter_to_letter, \
        get_entire_line, get_line_from_start_to_char, get_line_from_char_to_end, get_line_from_char_to_char

    # rendering (basic, highlighting, syntax coloring)
    from ._rendering import render_line_contents_by_dicts, \
        render_caret, caret_within_texteditor, render_background_coloring, render_line_numbers, \
        reset_text_area_to_caret, get_rect_coord_from_indizes, get_rect_coord_from_mouse
    from ._rendering_highlighting import render_highlight, highlight_lines, highlight_entire_line, \
        highlight_from_letter_to_letter, highlight_from_start_to_letter, highlight_from_letter_to_end
    from ._rendering_syntax_coloring import get_syntax_coloring_dicts, get_single_color_dicts

    from ._editor_getters import get_line_index, get_letter_index, line_is_visible, get_showable_lines, \
        get_number_of_letters_in_line_by_mouse, get_number_of_letters_in_line_by_index

    from ._other import jump_to_start, jump_to_end, reset_after_highlight

    # files for customization of the editor:
    from ._customization import set_line_numbers, set_syntax_highlighting, set_colorscheme, set_colorscheme_from_yaml
    from ._usage import get_text_as_list, get_text_as_string, clear_text, set_text_from_list, set_text_from_string

    def __init__(self, offset_x, offset_y, text_area_width, text_area_height, screen,
                 line_numbers_flag=False, style='dark', syntax_highlighting_flag=False):
        self.screen = screen

        # VISUALS
        self.editor_offset_X = offset_x
        self.editor_offset_Y = offset_y
        self.textAreaWidth = text_area_width
        self.textAreaHeight = text_area_height
        self.conclusionBarHeight = 18
        self.letter_size_Y = 15
        self.letter_size_X = 9
        current_dir = os.path.dirname(__file__)
        self.courier_font = pygame.font.Font(os.path.join(current_dir, "elements/fonts/Courier.ttf"),
                                             self.letter_size_Y)
        self.trennzeichen_image = pygame.image.load(
            os.path.join(current_dir, "elements/graphics/Trennzeichen.png")).convert_alpha()
        self.syntax_coloring = syntax_highlighting_flag

        # LINES
        self.Trenn_counter = 0
        self.MaxLinecounter = 0
        self.line_string_list = []  # LOGIC: Array of actual Strings
        self.lineHeight = 18
        self.showable_line_numbers_in_editor = int(math.floor(self.textAreaHeight / self.lineHeight))

        # self.maxLines is the variable keeping count how many lines we currently have -
        # in the beginning we fill the entire editor with empty lines.
        self.maxLines = int(math.floor(self.textAreaHeight / self.lineHeight))
        self.showStartLine = 0  # first line (shown at the top of the editor) <- must be zero during init!
        self.line_gap = 3 + self.letter_size_Y

        for i in range(self.maxLines):  # from 0 to maxLines:
            self.line_string_list.append("")  # Add a line

        # SCROLLBAR
        self.scrollBarImg = pygame.image.load(os.path.join(current_dir, "elements/graphics/Scroll_Bar.png")).convert()
        self.scrollBarWidth = 8  # must be an even number

        self.scrollbar: pygame.Rect = None
        self.scroll_start_y: int = None
        self.scroll_dragging: bool = False

        # LINE NUMBERS
        self.displayLineNumbers = line_numbers_flag
        if self.displayLineNumbers:
            self.lineNumberWidth = 27  # line number background width and also offset for text!
        else:
            self.lineNumberWidth = 0
        self.line_numbers_Y = self.editor_offset_Y

        # TEXT COORDINATES
        self.chosen_LineIndex = 0
        self.chosen_LetterIndex = 0
        self.yline_start = self.editor_offset_Y + 3
        self.yline = self.editor_offset_Y
        self.xline_start_offset = 28
        if self.displayLineNumbers:
            self.xline_start = self.editor_offset_X + self.xline_start_offset
            self.xline = self.editor_offset_X + self.xline_start_offset
        else:
            self.xline_start = self.editor_offset_X
            self.xline = self.editor_offset_X

        # CURSOR - coordinates for displaying the caret while typing
        self.cursor_Y = self.yline_start - 3
        self.cursor_X = self.xline_start

        # click down - coordinates used to identify start-point of drag
        self.dragged_active = False
        self.dragged_finished = True
        self.drag_chosen_LineIndex_start = 0
        self.drag_chosen_LetterIndex_start = 0
        self.last_clickdown_cycle = 0

        # click up  - coordinates used to identify end-point of drag
        self.drag_chosen_LineIndex_end = 0
        self.drag_chosen_LetterIndex_end = 0
        self.last_clickup_cycle = 0

        # Colors
        self.codingBackgroundColor = (40, 41, 41)
        self.codingScrollBarBackgroundColor = (49, 50, 50)
        self.lineNumberColor = (255, 255, 255)
        self.lineNumberBackgroundColor = (60, 61, 61)
        self.textColor = (255, 255, 255)
        self.color_scrollbar = (60, 61, 61)

        self.lexer = PythonLexer()
        self.formatter = PybrainzzFormatter()
        self.set_colorscheme(style)

        # Performance enhancing variables
        self.deleteCounter = 0
        self.firstiteration_boolean = True
        self.rerenderLineNumbers = True
        self.click_hold = False

        self.cycleCounter = 0  # Used to be able to tell whether a mouse-drag action has been handled already or not.

        self.clock = pygame.time.Clock()
        self.FPS = 60  # we need to limit the FPS so we don't trigger the same actions too often (e.g. deletions)

    def display_editor(self, pygame_events, pressed_keys, mouse_x, mouse_y, mouse_pressed):
        # needs to be called within a while loop to be able to catch key/mouse input and update visuals throughout use.
        self.cycleCounter = self.cycleCounter + 1
        # first iteration
        if self.firstiteration_boolean:
            # paint entire area to avoid pixel error beneath line numbers
            pygame.draw.rect(self.screen, self.codingBackgroundColor,
                             (self.editor_offset_X, self.editor_offset_Y, self.textAreaWidth, self.textAreaHeight))

            self.firstiteration_boolean = False

        for event in pygame_events:  # handle QUIT operation
            if event.type == pygame.QUIT:
                pygame.quit()
                exit()

        self.handle_keyboard_input(pygame_events, pressed_keys)
        self.handle_mouse_input(pygame_events, mouse_x, mouse_y, mouse_pressed)

        # RENDERING 1 - Background objects
        self.render_background_coloring()
        self.render_line_numbers()

        # RENDERING 2 - Lines
        self.render_highlight(mouse_x, mouse_y)

        if self.syntax_coloring:  # syntax highlighting for code
            list_of_dicts = self.get_syntax_coloring_dicts()
        else:  # single-color text
            list_of_dicts = self.get_single_color_dicts()
        self.render_line_contents_by_dicts(list_of_dicts)

        self.render_caret()

        self.render_scrollbar_vertical()
        self.clock.tick(self.FPS)



import pygame


def render_highlight(self, mouse_x, mouse_y) -> None:
    """
    Renders highlighted area:
    1. During drag-action -> area starts at drag_start and follows mouse
    2. After drag-action -> area stays confined to selected area by drag_start and drag_end
    """

    if self.dragged_active:  # some text is highlighted or being highlighted
        line_start = self.drag_chosen_LineIndex_start
        letter_start = self.drag_chosen_LetterIndex_start

        if self.dragged_finished:  # highlighting operation is done, user "clicked-up" with the left mouse button
            line_end = self.drag_chosen_LineIndex_end
            letter_end = self.drag_chosen_LetterIndex_end
            if letter_end < 0:
                letter_end = 0
            self.highlight_lines(line_start, letter_start, line_end, letter_end)  # Actual highlighting

        else:  # active highlighting -> highlighted area follows mouse movements
            line_end = self.get_line_index(mouse_y)
            letter_end = self.get_letter_index(mouse_x)
            # adapt line_end: if mouse_y below showable area / existing lines,
            if line_end >= self.get_showable_lines():
                line_end = self.get_showable_lines() - 1  # select last showable/existing line as line_end

            # Correct letter_end based on cursor position / letters in the cursor's line
            if letter_end < 0:  # cursor is left of the line
                letter_end = 0
            elif letter_end > len(self.line_string_list[line_end]):
                letter_end = len(self.line_string_list[line_end])

            self.highlight_lines(line_start, letter_start, line_end, letter_end)  # Actual highlighting


def highlight_lines(self, line_start, letter_start, line_end, letter_end) -> None:
    """
    Highlights multiple lines based on indizies of starting & ending lines and letters.
    """
    if line_start == line_end:  # single-line highlight
        self.highlight_from_letter_to_letter(line_start, letter_start, letter_end)
    else:  # multi-line highlighting
        if line_start > line_end:  # swap variables based on up/downward highlight to make code more readable
            line_start, line_end = line_end, line_start
            letter_start, letter_end = letter_end, letter_start

        for i, line_number in enumerate(range(line_start, line_end + 1)):  # for each line
            if i == 0:  # first line
                self.highlight_from_letter_to_end(line_number, letter_start)  # right leaning highlight
            elif i < len(range(line_start, line_end)):  # middle line
                self.highlight_entire_line(line_number)
            else:  # last line
                self.highlight_from_start_to_letter(line_number, letter_end)  # left leaning highlight


def highlight_from_letter_to_end(self, line, letter) -> None:
    """
    Highlight from a specific letter by index to the end of a line.
    """
    if self.line_is_visible(line):
        x1, y1 = self.get_rect_coord_from_indizes(line, letter)
        x2, y2 = self.get_rect_coord_from_indizes(line, len(self.line_string_list[line]))
        pygame.draw.rect(self.screen, (0, 0, 0), pygame.Rect(x1, y1, x2 - x1, self.lineHeight))


def highlight_from_start_to_letter(self, line, letter) -> None:
    """
    Highlight from the beginning of a line to a specific letter by index.
    """
    if self.line_is_visible(line):
        x1, y1 = self.get_rect_coord_from_indizes(line, 0)
        x2, y2 = self.get_rect_coord_from_indizes(line, letter)
        pygame.draw.rect(self.screen, (0, 0, 0), pygame.Rect(x1, y1, x2 - x1, self.lineHeight))


def highlight_entire_line(self, line) -> None:
    """
    Full highlight of the entire line - first until last letter.
    """
    if self.line_is_visible(line):
        x1, y1 = self.get_rect_coord_from_indizes(line, 0)
        x2, y2 = self.get_rect_coord_from_indizes(line, len(self.line_string_list[line]))
        pygame.draw.rect(self.screen, (0, 0, 0), pygame.Rect(x1, y1, x2 - x1, self.lineHeight))


def highlight_from_letter_to_letter(self, line, letter_start, letter_end) -> None:
    """
    Highlights within a single line from letter to letter by indizes.
    """
    if self.line_is_visible(line):
        x1, y1 = self.get_rect_coord_from_indizes(line, letter_start)
        x2, y2 = self.get_rect_coord_from_indizes(line, letter_end)
        pygame.draw.rect(self.screen, (0, 0, 0), pygame.Rect(x1, y1, x2 - x1, self.lineHeight))



from pygments import highlight
from pygments.lexers import PythonLexer
from pygments.formatter import Formatter
from pygments.token import Token, is_token_subtype
from typing import List, Dict, Tuple
import os
import yaml
import re


class PybrainzzFormatter:

    def __init__(self):
        self.textColor_normal = (0, 0, 0)
        self.textColor_comments = (0, 0, 0)
        self.textColor_quotes = (0, 0, 0)
        self.textColor_operators = (0, 0, 0)
        self.textColor_keywords = (0, 0, 0)
        self.textColor_function = (0, 0, 0)
        self.textColor_builtin = (0, 0, 0)

    def format(self, tokensource: List[Tuple]) -> List[Dict]:
        """
        Input is a list of tuples. Each tuple contains the Token and the corresponding characters of said token.
        For different types of tokens the function creates a dict with the three keys:
        - chars -> String
        - type -> String
        - color (rgb color -> (0,0,0) )
        Depending on the type of token a color is assigned.
        """

        dicts = []
        for ttype, value in tokensource:
            # NAME.FUNCTION
            if ttype == Token.Name.Function:
                dicts.append({'chars': value, 'type': 'function', 'color': self.textColor_function})

            # BUILTIN
            elif ttype == Token.Name.Builtin:
                dicts.append({'chars': value, 'type': 'builtin', 'color': self.textColor_builtin})

            # LITERAL.STRING(.*)
            elif ttype == Token.Literal.String.Affix:
                dicts.append({'chars': value, 'type': 'builtin', 'color': self.textColor_builtin})
            elif is_token_subtype(ttype, Token.Literal.String):
                dicts.append({'chars': value, 'type': 'quotes', 'color': self.textColor_quotes})

            # OPERATOR.*
            elif is_token_subtype(ttype, Token.Operator):
                dicts.append({'chars': value, 'type': 'operators', 'color': self.textColor_operators})

            # KEYWORD.*
            elif is_token_subtype(ttype, Token.Keyword):
                dicts.append({'chars': value, 'type': 'keywords', 'color': self.textColor_keywords})

            # COMMENT.*
            elif is_token_subtype(ttype, Token.Comment):
                dicts.append({'chars': value, 'type': 'comments', 'color': self.textColor_comments})

            elif value == '\n':  # since we parse line-by-line, we need to exclude the linebreak character
                pass

            # REST
            else:
                dicts.append({'chars': value, 'type': 'text', 'color': self.textColor_normal})

        return dicts
